from pulsar.async.consts import ACTOR_STATES
from pulsar import send
import asyncio
from celebi.core.types import MaybeAsyncCallable
from asyncio import iscoroutinefunction


class Measurement():

    def __init__(
            self,
            name,
            spout: MaybeAsyncCallable,
            rate: int
    ):
        self.name = name
        self.spout = spout
        self.rate = rate

    def __aiter__(self):
        return self

    async def __anext__(self):
        await asyncio.sleep(self.rate)
        if iscoroutinefunction(self.spout):
            return await self.spout()
        else:
            return self.spout()

    async def measure(self, actor, exc=None):
        async for datum in self:
            if actor.state is ACTOR_STATES.STOPPING:
                return
            await send(
                actor.monitor,
                'fire',
                'spout',
                data=datum
            )



from pulsar import get_event_loop
from celebi.settings import POSTGRES
import asyncpg


def split_file(filename: str) -> list:
    with open(filename, 'r') as f:
        data = ''.join(map(lambda l: l.strip(), f.readlines()))
        return [q for q in data.split(';') if q]


async def async_init_db(config=POSTGRES):
    drop_sqls = split_file('celebi/schema/drop.sql')

    create_sqls = split_file('celebi/schema/schema.sql')
    conn = await asyncpg.connect(**config)
    res = [await conn.execute(sql) for sql in drop_sqls + create_sqls]

    return res


def init_db(config=POSTGRES):
    loop = get_event_loop()
    return loop.run_until_complete(async_init_db(config))



# -*- eval: (venv-workon "celebi"); -*-

from typing import Union, Callable, Generic, GenericMeta, TypeVar

from types import CoroutineType
from pulsar.apps.wsgi import WsgiResponse, WsgiRequest

__all__ = ['Request', 'Response',
           'MaybeCorouteine', 'Handler',
           'MaybeType', 'Maybe']


Request = WsgiRequest
Response = WsgiResponse

MaybeCorouteine = Union[CoroutineType, Callable]
Handler = Callable[..., WsgiResponse]


MaybeType = Generic[TypeVar("Just"), TypeVar("Error")]
Maybe: GenericMeta = type("Maybe", (MaybeType, ), {})

AsyncCallable = Callable[..., CoroutineType]
MaybeAsyncCallable = Union[AsyncCallable, Callable]



from typing import NamedTuple
from typing import NewType
import datetime

__all__ = ['json', 'varchar',
           'boolean', 'Table']

json = NewType('json', dict)
varchar = NewType('varchar', str)
boolean = NewType('boolean', bool)
text = NewType('text', str)
integer = NewType('integer', int)
timestamp = NewType('timestamp', datetime.datetime)
array = NewType('array', list)
Table = NamedTuple



# -*- eval: (venv-workon "celebi"); -*-

from typing import Callable, Any
import json
from functools import wraps
from asyncio import coroutine
from celebi.core.wsgi import success_response, error_response
from celebi.core.types import (MaybeCorouteine, Handler,
                               Response, Maybe)
from asyncio import iscoroutinefunction
import traceback


__all__ = ['maybe_coroutine', 'maybe_async_cps', 'jsonrpc']


def maybe_coroutine(fn: Callable) -> Callable[..., MaybeCorouteine]:
    if iscoroutinefunction(fn):
        return coroutine
    else:
        return lambda x: x


def maybe_async_cps(fn: Callable) -> MaybeCorouteine:

    def parted(context: Callable[[Maybe], Response]):
        @wraps(fn)
        @maybe_coroutine(fn)
        def _(*args, **kwargs):
            try:
                if iscoroutinefunction(fn):
                    result = yield from fn(*args, **kwargs)
                else:
                    result = fn(*args, **kwargs)
            except Exception as e:
                result = e
            return context(result)
        return _
    return parted


def jsonrpc(fn: Callable) -> Handler:
    def error(e: Exception) -> Response:
        try:
            raise e
        except Exception:
            traceback.print_exc()
        finally:
            res = dict(result=None, error=e.args, id=1)
            return error_response(json.dumps(res).encode())

    def just(r: dict) -> Response:
        res = dict(result=r, error=None, id=1)
        return success_response(json.dumps(res).encode())

    @maybe_async_cps(fn)
    def handler(res: Maybe[Any, Exception]) -> Response:
        if isinstance(res, Exception):
            return error(res)
        else:
            return just(res)

    return handler



import asyncpg
from pulsar.apps import Application
from pulsar import send, get_actor, get_application

__all__ = ['PostgresMonitor']


class PostgresMonitor(Application):
    name = "postgres_arbiter"

    async def _connect(self):
        return await asyncpg.connect(**self.cfg.pgconfigs)

    @staticmethod
    async def _execute(actor, sql) -> str:
        return str(
            await actor.conn.execute(sql)
        )

    @staticmethod
    async def _fetchrow(actor, sql) -> dict:
        return dict(
            await actor.conn.fetchrow(sql)
        )

    @staticmethod
    async def _fetch(actor, sql) -> dict:
        return list(
            map(dict, await actor.conn.fetch(sql))
        )

    @staticmethod
    async def _transaction(actor, sqls: list):
        async with actor.conn.transaction():
            return [
                await actor.conn.execute(sql)
                for sql in sqls
            ]

    async def monitor_start(self, monitor, exc=None):
        monitor.conn = await self._connect()

    async def worker_start(self, worker, exc=None):
        worker.conn = await self._connect()

    @classmethod
    async def query(cls, sql):
        arbiter = get_actor().get_actor(cls.name)
        app = await get_application(cls.name)
        actor = await arbiter.spawn(
            start=app.worker_start
        )
        return await send(actor, 'run', app._fetch, sql)

    @classmethod
    async def update(cls, sql):
        arbiter = get_actor().get_actor(cls.name)
        app = await get_application(cls.name)
        actor = await arbiter.spawn(
            start=app.worker_start
        )
        return await send(actor, 'run', app._fetch, sql)

    @classmethod
    async def insert(cls, sql):
        arbiter = get_actor().get_actor(cls.name)
        app = await get_application(cls.name)
        actor = await arbiter.spawn(
            start=app.worker_start
        )
        return await send(actor, 'run', app._fetchrow, sql)

    @classmethod
    async def execute(cls, sql: str) -> str:

        arbiter = get_actor().get_actor(cls.name)
        app = await get_application(cls.name)
        actor = await arbiter.spawn(
            start=app.worker_start
        )
        return await send(actor, 'run', app._execute, sql)

    @classmethod
    async def transaction(cls, sqls: list) -> str:
        arbiter = get_actor().get_actor(cls.name)
        app = await get_application(cls.name)
        actor = await arbiter.spawn(
            start=app.worker_start
        )
        return await send(actor, 'run', app._execute, sqls)



# -*- eval: (venv-workon "celebi"); -*-

from functools import partial
from typing import Callable
from pwsgi.router import BluePrint
from pulsar.apps.wsgi import WsgiResponse


__all__ = ['wsgi', 'error_response', 'success_response']


wsgi: BluePrint = BluePrint('/')
success_response: Callable = partial(WsgiResponse, 200)
error_response: Callable = partial(WsgiResponse, 500)

response = success_response



from pulsar.apps import ws
from pulsar import ensure_future
import logging
import json
from pulsar.apps.wsgi import route

__all__ = ['CelebiSocket']


class WebSocketRouter(ws.WebSocket):

    @route('/<int:qid>/', method='get')
    def qubit(self, request):
        qid = list(filter(bool, (request.path.split('/'))))[-1]
        request.qid = qid
        return super().get(request)


class PubSubWS(ws.WS):
    '''A :class:`.WS` handler with a publish-subscribe handler
    '''
    client = ws.PubSubClient

    def __init__(self):
        pass

    def on_open(self, websocket):
        qid = websocket.handshake.qid
        channel = self.channel % qid
        self.pubsub.add_client(self.client(websocket, channel))
        ensure_future(self.pubsub.subscribe(channel))
        logging.info(
            'New websocket opened. Add client to %s on "%s" channel',
            self.pubsub, self.channel)

    def write(self, websocket, message):
        data = json.loads(message)
        if not str(data['qid']) == str(websocket.handshake.qid):
            return
        logging.info('qid: %s, data %s' % (websocket.handshake.qid, data))
        websocket.write(data['data'])


CelebiSocket = WebSocketRouter(
    '/subscribe', PubSubWS())



import asyncio
from typing import Any
from collections import defaultdict
import traceback

__all__ = ['configdict']


def configdict(cases: dict, other: Any) -> defaultdict:
    assert other in cases.keys(), 'cases should include default cases'
    return defaultdict(lambda: other)(cases)


def retry(fn, times=3):
    async def _(*args, **kwargs):
        for t in range(0, times):
            await asyncio.sleep(t * 5)
            try:
                return await fn(*args, **kwargs)
            except Exception as e:
                if t + 1 != times:
                    traceback.print_exc()
                    continue
                raise(e)
            print('Retry %snd time, Waiting %s seconds' % (t + 1, (t + 1) * 5))
    return _


def ensure(fn):
    def _(*args, **kwargs):
        return asyncio.ensure_future(fn(*args, **kwargs))
    return _



import asyncio
import pika
from pika.connection import URLParameters
from pulsar import send
from pulsar.async.consts import ACTOR_STATES
from celebi.utils import retry


class Entanglement():

    def __init__(
            self,
            name: str,
            rate: int,
            exchange: str,
            amqp_url,
            mapper=lambda x: x,
            *args,
            **kwargs
    ):
        self.name = name
        self.rate = rate
        self.exchange = exchange
        self.amqp_url = amqp_url

    async def mapper(self, datum):
        return datum

    @retry
    async def connect(self):
        url = URLParameters(self.amqp_url)
        self.connection = pika.BlockingConnection(url)
        self.channel = self.connection.channel()
        self.queue = self.channel.queue_declare(exclusive=True).method.queue

    async def handler(self, channel, method, properties, body):
        await send(self.actor.monitor, 'fire', 'spout', body)

    async def entangle(self, actor, exc=None):
        await self.connect()
        self.channel.queue_bind(
            exchange=self.exchange,
            queue=self.queue
        )
        # Check pika.adepter.process_data_events
        while actor.state is not ACTOR_STATES.STOPPING:
            await asyncio.sleep(self.rate)
            body: bytes = self.channel.basic_get(self.queue)[2]
            if not body:
                continue
            datum = await self.mapper(body.decode())
            await send(
                actor.monitor,
                'fire',
                'spout',
                data=datum
            )
        self.connection.close()



# -*- eval: (venv-workon "celebi"); -*-

from .wsgi import wsgi, success_response, error_response, response
from .decorators import jsonrpc

__all__ = ['wsgi',
           'success_response',
           'error_response',
           'response',
           'jsonrpc']



#! -*- eval: (venv-workon "celebi"); -*-
import time
import json
from celebi.io.postgres import types
from celebi.io.postgres import QuerySet


__all__ = ['Data', 'Datum']


class Data(object):
    prototype = types.Table('ts_data', [
        ('id', types.integer),
        ('name', types.varchar),
        ('comment', types.text),
    ])
    manager = QuerySet(prototype)

    @classmethod
    async def create(cls, *args, **kwargs) -> str:
        return await cls.manager.insert(*args, **kwargs)

    @classmethod
    async def fetch(cls, did) -> str:
        return await cls.manager.get_by(id=did)

    @classmethod
    async def delete(cls, did) -> str:
        return await cls.manager.delete(did)

    @classmethod
    async def update(cls, did, data) -> str:
        return await cls.manager.update(did, **data)


class Datum(object):
    prototype = types.Table('ts_datum', [
        ('id', types.integer),
        ('ts', types.timestamp),
        ('dataset', types.integer),
        ('datum', types.json),
        ('tags', types.text)
    ])
    manager = QuerySet(prototype)

    @classmethod
    async def create(cls, dataset, datum, tags=[], ts=time.time()):
        return await cls.manager.insert(
            dataset=dataset,
            datum=json.dumps(datum),
            tags=','.join(tags),
            ts=ts
        )

    @classmethod
    async def fetch(cls, did) -> str:
        return await cls.manager.get_by(id=did)

    @classmethod
    async def delete(cls, did) -> str:
        return await cls.manager.delete(did)

    @classmethod
    async def update(cls, did, data) -> str:
        return await cls.manager.update(did, **data)



# -*- eval: (venv-workon "celebi"); -*-
import os

ENV = os.getenv('ENV') or 'TEST'

POSTGRES_TEST = {
    'user': 'postgres',
    'password': '',
    'host': 'postgres',
    'port': 5432
}


POSTGRES_DEV = {
    'user': 'postgres',
    'password': '',
    'host': '127.0.0.1',
    'port': 32770
}

POSTGRES = {
    'TEST': POSTGRES_TEST,
    'DEV': POSTGRES_DEV
}.get(ENV, POSTGRES_TEST)

RABBITMQ_TEST = "amqp://guest:guest@rabbitmq:5672/%2F"
RABBITMQ_DEV = "amqp://guest:guest@localhost:32773/%2F"
RABBITMQ = {
    'TEST': RABBITMQ_TEST,
    'DEV': RABBITMQ_DEV
}.get(ENV, RABBITMQ_TEST)



# -*- eval: (venv-workon "celebi"); -*-

from pulsar.apps import MultiApp
from functools import partial
from pulsar.apps.wsgi.handlers import WsgiHandler
from pulsar.apps.wsgi import WSGIServer
from celebi.settings import POSTGRES, RABBITMQ
from celebi.core import wsgi
from celebi.io.postgres import PostgresMonitor
from celebi import apis
from celebi.io.qubit import (
    Measurement,
    Entanglement,
    QubitMonitor
)

__all__ = ['apis', 'wsgi', 'ComposedApp', 'ComposedIO']


class ComposedIO(MultiApp):
    name = 'arbiters'

    def build(self):
        yield self.new_app(
            App=QubitMonitor,
            name='pikachu_monitor',
            exchange_type='fanout',
            exchange='test',
            processor=dict,
            measurements=[Measurement(
                name='test',
                spout=partial(dict, hello='world'),
                rate=1
            )],
            entanglements=[
                Entanglement(
                    name='test',
                    exchange='test',
                    rate=1,
                    amqp_url=RABBITMQ
                )
            ],
            amqp_cfg=RABBITMQ,
            workers=10
        )
        yield self.new_app(
            App=PostgresMonitor,
            workers=10,
            pgconfigs=POSTGRES
        )


class ComposedApp(MultiApp):
    name = 'celebi'

    def build(self):
        yield self.new_app(WSGIServer, callable=WsgiHandler((wsgi, )), bind="127.0.0.1:8964")
        yield self.new_app(ComposedIO)



import os

from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))


install_requires = [
    line
    for line in open(
        os.path.join(here, "requirements.txt"),
        "r"
    )
]
author = 'ryankung'
email = 'ryankung@ieee.org'


setup(
    name='celebi',
    description='celebi is a Actor Based TS data anaylize engine',  # noqa
    version='0.0.1',
    packages=find_packages(here, exclude=['tests']),
    license='GPL',
    author=author,
    author_email=email,
    install_requires=install_requires,
)



def wrap_key(x: str) -> str:
    return '%s' % str(x)


def wrap_value(x: str) -> str:
    return "'%s'" % str(x)


def concat(xs: list) -> str:
    return ','.join(xs)


def escape(x: str) -> str:
    x = str(x)
    x.replace("'", '"')
    return x


def get_and_seg(d: dict) -> str:
    return ' and '.join("%s='%s'" % (str(k), str(v)) for k, v in d.items())


def get_pairs(dct):
    return ' , '.join("%s='%s'" % (str(k), str(v)) for k, v in dct.items())



import pika
from asyncio import iscoroutinefunction
from pika.connection import URLParameters
import json
from pulsar.apps import Application
from .entangle import Entanglement
from .measure import Measurement
from pulsar import command
from pulsar.async.consts import ACTOR_STATES
from celebi.utils import retry, ensure

__all__ = ['QubitMonitor']


@command(ack=False)
async def fire(request, event, **kw):
    if request.actor.state is not ACTOR_STATES.STOPPING:
        request.actor.fire_event(event, **kw)


class QubitMonitor(Application):
    def __init__(
            self,
            cfg,
            *args,
            **kwargs
    ):
        self.name = cfg.name
        self.exchange: str = cfg.exchange
        self.exchange_type: str = cfg.exchange_type
        self._measurements = cfg.measurements
        self._measuring = []
        self._entanglements = cfg.entanglements
        self._entangleing = []
        self._processor = cfg.processor
        super().__init__(cfg=cfg, *args, **kwargs)

    @ensure
    async def spout(
            self,
            monitor,
            data,
            **kwargs
    ):
        if type(data) is str:
            data = json.loads(data)

        if iscoroutinefunction(self._processor):
            data = await self._processor(data)
        else:
            data = self._processor(data)
        assert self.channel.basic_publish(
            exchange=self.exchange,
            routing_key='',
            body=json.dumps(data)
        )
        return

    @retry
    async def connect_channel(self):
        url = URLParameters(self.cfg.amqp_cfg)
        self.connection = pika.BlockingConnection(url)
        self.channel = self.connection.channel()
        self.channel.exchange_declare(
            exchange=self.exchange,
            exchange_type=self.exchange_type
        )

    async def monitor_start(self, monitor, exc=None):
        monitor.bind_event('spout', self.spout)
        await self.connect_channel()
        self._measuring: list = [
            self.spawn_measuring_actor(monitor, m)
            for m in self._measurements
        ]
        self._entangling: list = [
            self.spawn_entangling_actor(monitor, e)
            for e in self._entanglements
        ]

    async def monitor_stopping(self, monitor, exc=None):
        self.connection.close()

    def spawn_measuring_actor(self, monitor, m: Measurement):
        actor = monitor.spawn(
            name=m.name,
            start=m.measure
        )
        return actor

    def spawn_entangling_actor(self, monitor, e: Entanglement):
        actor = monitor.spawn(
            name=e.name,
            start=e.entangle
        )
        return actor

    def measure(self, m: Measurement):
        self._measurements.append(m)

    def entangle(self, e: Entanglement):
        self._entanglements.append(e)



from .monitor import PostgresMonitor
from . import utils
import time

__all__ = ['QuerySet']

key = str(time.time())


class QuerySet(object):
    _sql = {
        'get_list': 'SELECT {fields} from {table} {extra} LIMIT {size} OFFSET {offset}',
        'filter': 'SELECT {fields} from {table} WHERE {rule} LIMIT {size} OFFSET {offset}',
        'count': 'SELECT COUNT({field}) FROM {table}',
        'count_on_rule': 'SELECT COUNT({field}) FROM {table} WHERE {rule}',
        'orderby': 'ORDER BY {field}',
        'nearby': 'select {fields} difference from {table} where {rule} and {value} > {column} limit 1',
        'orderby_decr': 'ORDER BY {field} DECR',
        'filter_with_orderby': "SELECT {fields} from {table} WHERE {rule} ORDER BY {sort_key} LIMIT {size} OFFSET {offset};",
        'filter_with_orderby_decr': "SELECT {fields} from {table} WHERE {rule} ORDER BY {sort_key} LIMIT {size} OFFSET {offset};",
        'filter_in': "SELECT {fields} FROM {table} WHERE {key} IN ({targets});",
        'filter_in_range': "SELECT {fields} FROM {table} WHERE {rule} and {key} <= {end} and {key} >= {start};",
        'find_in_range': "SELECT {fields} FROM {table} WHERE {key} <= {end} and {key} >= {start};",
        'find_near': "SELECT {fields} FROM {table} WHERE {key} >= {start};",
        'insert': 'INSERT INTO {table} ({keys}) VALUES ({values}) RETURNING id;',
        'replace': 'REPLACE INTO {table} ({keys}) VALUES ({values})',
        'delete': "DELETE FROM {table} WHERE {rules} RETURNING id",
        'update': "UPDATE {table} SET {key_value_pairs} WHERE {rules} RETURNING id",
        'append_array': "UPDATE {table} SET {key} = array_append({key}, {value}) WHERE id='{id}' RETURNING id",
        'get_via_id': "SELECT {fields} from {table} WHERE id='{id}'",
        'update_via_id': "UPDATE {table} SET {key_value_pairs} WHERE id='{id}' RETURNING id",
        'delete_via_id': "DELETE FROM {table} WHERE id='{id}' RETURNING id",
        'incr': "UPDATE {table} SET {key}={key}+'{num}' WHERE id='{id}' RETURNING id",
        'decr': "UPDATE {table} SET {key}={key}-'{num}' WHERE id='{id}' RETURNING id",
        'search': "SELECT {fields} FROM {table} WHERE {extra} {key} LIKE '%{value}%' LIMIT {size} OFFSET {offset}",
        'insert_or_update': "INSERT INTO {table} ({keys}) VALUES ({values}) ON DUPLICATE KEY UPDATE {key_value_pairs};"
    }

    def __init__(self, table):
        self.table = table
        self.fields = table._fields
        self.tablename = table.__name__
        self.db = PostgresMonitor

    def format(self, data):
        if not isinstance(data, dict):
            return utils.escape(str(data.encode('utf8')))
        if not all(f in self.fields for f in data.keys()):
            raise Exception("Unknew Fields", set(
                data.keys()) - set(self.fields))
        try:
            res = {k: utils.escape(v) for k, v in data.items()}
            return res
        except:
            raise Exception("Series Failed")

    async def nearby(self, value, column, *args, **kwargs):
        data = self.format(kwargs)
        return await self.db.query(self._sql['nearby'].format(**{
            'table': self.tablename,
            'fields': utils.concat(map(utils.wrap_key, self.fields)),
            'value': utils.escape(value),
            'column': utils.escape(column),
            'rule': utils.get_and_seg(data)
        }))

    async def get(self, oid):
        res = await self.db.query(self._sql['get_via_id'].format(**{
            'table': self.tablename,
            'fields': utils.concat(map(utils.wrap_key, self.fields)),
            'id': oid
        }))
        return res and dict(res) if res else None

    async def get_by(self, *args, **kwargs):
        data = self.format(kwargs)
        res = await self.db.query(self._sql['filter'].format(**{
            'table': self.tablename,
            'rule': utils.get_and_seg(data),
            'size': '1',
            'offset': '0',
            'fields': utils.concat(map(utils.wrap_key, self.fields)),
        }))
        return res and dict(res[0])

    async def search(self, key, value, start, limit, filters=''):
        return await self.db.query(self._sql['search'].format(**{
            'table': self.tablename,
            'fields': utils.concat(map(utils.wrap_key, self.fields)),
            'key': self.format(key),
            'value': self.format(value),
            'offset': str(int(start)),
            'size': str(int(limit)),
            'extra': filters and utils.get_pairs(filters) + 'and' or ''
        }))

    async def get_list(self, size=100, offset=0, sort_key='') -> list:
        if isinstance(sort_key, list):
            sort_key = utils.concat(map(utils.set_desc, sort_key))
        else:
            sort_key = sort_key and utils.set_desc(sort_key) or ''

        res = await self.db.query(self._sql['get_list'].format(**{
            'table': self.tablename,
            'fields': utils.concat(map(utils.wrap_key, self.fields)),
            'size': str(int(size)),
            'offset': str(int(offset)),
            'extra': sort_key and self._sql['orderby'].format(**{
                'field': sort_key
            }) or ''

        }))
        return [dict(r) for r in res]

    async def find_in(self, key, targets, fields=[]) -> dict:
        return await self.db.query(self._sql['filter_in'].format(**{
            'table': self.tablename,
            'fields': utils.concat(map(utils.wrap_key, fields or self.fields)),
            'key': key,
            'targets': utils.concat(map(utils.wrap_value, targets))
        }))

    async def find_near(self, key, start, end, fields=[], *args, **kwargs) -> dict:
        data = self.format(kwargs)
        res = await self.db.query(self._sql['find_near'].format(**{
            'table': self.tablename,
            'fields': utils.concat(map(utils.wrap_key, fields or self.fields)),
            'key': key,
            'rule': utils.get_and_seg(data),
            'start': utils.wrap_value(start),
            'end': utils.wrap_value(end)
        }))
        return [dict(r) for r in res]

    async def find_in_range(self, key, start, end, fields=[], *args, **kwargs) -> dict:
        data = self.format(kwargs)
        res = await self.db.query(self._sql['filter_in_range'].format(**{
            'table': self.tablename,
            'fields': utils.concat(map(utils.wrap_key, fields or self.fields)),
            'key': key,
            'rule': utils.get_and_seg(data),
            'start': utils.wrap_value(start),
            'end': utils.wrap_value(end)
        }))
        return [dict(r) for r in res]

    async def count(self, field):
        field = utils.escape(field) or '*'
        return await self.db.query(self._sql['count'].format(**{
            'table': self.tablename,
            'field': field
        }))

    async def count_on_rule(self, field, rule):
        rule = self.format(rule)
        field = utils.escape(field)
        return await self.db.query(self._sql['count_on_rule'].format(**{
            'table': self.tablename,
            'rule': utils.get_and_seg(rule),
            'field': field
        }))

    async def filter(self, limit=100, offset=0, sort_key='', *args, **kwargs):
        data = self.format(kwargs)
        res = await self.db.query(self._sql['filter'].format(**{
            'table': self.tablename,
            'rule': utils.get_and_seg(data),
            'size': str(int(limit)),
            'fields': utils.concat(map(utils.wrap_key, self.fields)),
            'offset': str(int(offset))
        }))
        return [dict(r) for r in res]

    async def sortby(self, sort_key='id', offset=0, limit=100, extra="", decr=False, *args, **kwargs):
        data = self.format(kwargs)
        if isinstance(sort_key, list):
            sort_key = utils.concat(map(utils.set_desc, sort_key))
        else:
            sort_key = utils.set_desc(sort_key)
        tmpl = decr and 'filter_with_orderby_decr' or 'filter_with_orderby'
        return await self.db.query(self._sql[tmpl].format(**{
            'table': self.tablename,
            'rule': utils.get_and_seg(data),
            'size': str(int(limit)),
            'sort_key': sort_key,
            'offset': str(int(offset)),
            'fields': utils.concat(map(utils.wrap_key, self.fields)),
        }))

    async def insert(self, *args, **kwargs):
        data = self.format(kwargs)
        return await self.db.insert(self._sql['insert'].format(**{
            'table': self.tablename,
            'keys': utils.concat(map(utils.wrap_key, data.keys())),
            'values': utils.concat(map(utils.wrap_value, data.values()))
        }))

    async def replace(self, *args, **kwargs):
        data = self.format(kwargs)
        return await self.db.insert(self._sql['replace'].format(**{
            'table': self.tablename,
            'keys': utils.concat(map(utils.wrap_key, data.keys())),
            'values': utils.concat(map(utils.wrap_value, data.values()))
        }))

    async def update(self, oid, *args, **kwargs):
        data = self.format(kwargs)
        pairs = utils.get_pairs(data)
        return await self.db.update(self._sql['update_via_id'].format(**{
            'id': oid,
            'table': self.tablename,
            'key_value_pairs': pairs
        }))

    async def append_array(self, oid, key, value):
        return await self.db.update(self._sql['append_array'].format(**{
            'id': oid,
            'table': self.tablename,
            'key': key,
            'value': value
        }))

    async def insert_or_update(self, *args, **kwargs) -> dict:
        data = self.format(kwargs)
        return await self.db.insert(self._sql('insert_or_update').format(**{
            'table': self.tablename,
            'keys': utils.concat(map(utils.wrap_key, data.keys())),
            'values': utils.concat(map(utils.wrap_key, data.values())),
            'key_value_pairs': utils.get_pairs(data)
        }))

    async def update_by(self, rules, *args, **kwargs):
        data = self.format(kwargs)
        rules = self.format(rules)
        return await self.db.update(self._sql['update'].format(**{
            'table': self.tablename,
            'rules': utils.get_and_seg(rules),
            'key_value_pairs': utils.get_pairs(data)
        }))

    async def delete(self, oid):
        return await self.db.update(self._sql['delete_via_id'].format(**{
            'table': self.tablename,
            'id': oid
        }))

    async def delete_by(self, *args, **kwargs):
        data = self.format(kwargs)
        return await self.db.update(self._sql['delete'].format(**{
            'table': self.tablename,
            'rules': utils.get_and_seg(data)
        }))

    async def incr(self, oid, key, num):
        return await self.db.update(self._sql['incr'].format(**{
            'id': oid,
            'table': self.tablename,
            'key': key,
            'num': num
        }))

    async def decr(self, oid, key, num):
        return await self.db.update(self._sql['decr'].format(**{
            'id': oid,
            'table': self.tablename,
            'key': key,
            'num': num
        }))
